/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <dd-control.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/common.h"
#include "core/devices.h"
#include "extcon/extcon.h"

#define CONTROL_HANDLER_NAME		"control"
#define CONTROL_GETTER_NAME			"getcontrol"

static const struct control_device {
	const int id;
	const char *name;
} devices[] = {
	/* Add id & ops to provide start/stop control */
	{ DEVICE_CONTROL_MMC,			"mmc" },
	{ DEVICE_CONTROL_USBCLIENT,		"USB" },
	{ DEVICE_CONTROL_BLOCK,			"block" },
};

static int control_handler(int argc, char **argv)
{
	int i;
	int pid;
	int device;
	bool enable;
	int ret;
	const struct device_ops *dev_ops = NULL;

	_I("argc=%d", argc);
	for (i = 0; i < argc; ++i)
		_I("argv[%2d]=%s", i, argv[i]);

	pid = atoi(argv[0]);
	device = atoi(argv[1]);
	enable = atoi(argv[2]);
	_I("pid=%d device=%d enable=%d", pid, device, enable);

	for (i = 0; i < ARRAY_SIZE(devices); i++)
		if (devices[i].id == device)
			break;

	if (i >= ARRAY_SIZE(devices))
		return -EINVAL;

	FIND_DEVICE(dev_ops, devices[i].name);
	if (!check_default(dev_ops)) {
		_I("Device %s found.", devices[i].name);

		if (enable)
			ret = device_start(dev_ops);
		else
			ret = device_stop(dev_ops);
	} else {
		if (enable)
			ret = extcon_enable_device(devices[i].name);
		else
			ret = extcon_disable_device(devices[i].name);
	}

	return ret;
}

static int get_control_handler(int argc, char **argv)
{
	int i;
	int pid;
	int device;
	const struct device_ops *dev_ops = NULL;

	_I("argc=%d", argc);
	for (i = 0; i < argc; ++i)
		_I("argv[%2d]=%s", i, argv[i]);

	pid = atoi(argv[0]);
	device = atoi(argv[1]);
	_I("pid=%d device=%d", pid, device);

	for (i = 0; i < ARRAY_SIZE(devices); i++)
		if (devices[i].id == device)
			break;

	if (i >= ARRAY_SIZE(devices))
		return -EINVAL;

	FIND_DEVICE(dev_ops, devices[i].name);
	if (!check_default(dev_ops))
		return device_get_status(dev_ops);
	else
		return extcon_get_status(devices[i].name);
}

GVariant *dbus_control_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret;
	int argc;
	char *argv[3];

	g_variant_get(param, "(sisss)", NULL, &argc, &argv[0], &argv[1], &argv[2]);
	if (argc != 3) {
		_E("Invalid argument.");
		ret = -EINVAL;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("Process(%d) does not exist. Dbus ignored.", pid);
		ret = -ESRCH;
		goto out;
	}

	ret = control_handler(argc, (char **)&argv);
out:
	g_free(argv[0]);
	g_free(argv[1]);
	g_free(argv[2]);
	return g_variant_new("(i)", ret);
}


GVariant *dbus_get_control_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret;
	int argc = 0;
	char *argv[2];

	g_variant_get(param, "(siss)", NULL, &argc, &argv[0], &argv[1]);
	if (argc != 2) {
		_E("Invalid argument.");
		ret = -EINVAL;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("Process(%d) does not exist. Dbus ignored.", pid);
		ret = -ESRCH;
		goto out;
	}

	ret = get_control_handler(argc, (char **)&argv);
out:
	g_free(argv[0]);
	g_free(argv[1]);
	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ CONTROL_HANDLER_NAME, "sisss", "i", dbus_control_handler     },
	{ CONTROL_GETTER_NAME ,  "siss", "i", dbus_get_control_handler },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_SYSNOTI,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void control_init(void *data)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_SYSNOTI, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);

}

static const struct device_ops control_device_ops = {
	DECLARE_NAME_LEN("control"),
	.init     = control_init,
};

DEVICE_OPS_REGISTER(&control_device_ops)
